/*============================================================================
 *
 *     This file is part of the Code_Saturne Kernel, element of the
 *     Code_Saturne CFD tool.
 *
 *     Copyright (C) 1998-2016 EDF S.A., France
 *
 *     contact: saturne-support@edf.fr
 *
 *     The Code_Saturne Kernel is free software; you can redistribute it
 *     and/or modify it under the terms of the GNU General Public License
 *     as published by the Free Software Foundation; either version 2 of
 *     the License, or (at your option) any later version.
 *
 *     The Code_Saturne Kernel is distributed in the hope that it will be
 *     useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 *     of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with the Code_Saturne Kernel; if not, write to the
 *     Free Software Foundation, Inc.,
 *     51 Franklin St, Fifth Floor,
 *     Boston, MA  02110-1301  USA
 *
 *============================================================================*/

#ifndef _FSI_IDL_
#define _FSI_IDL_

#include "DSC_Engines.idl"
#include "SALOME_Exception.idl"
#include "SALOME_Component.idl"
#include "SALOME_Comm.idl"

module FSI_ORB
{
  interface FSI_ASTER: Engines::Superv_Component
  {
    void op0117(in string jdc) raises (SALOME::SALOME_Exception);
  };

  interface FSI_SATURNE: Engines::Superv_Component
  {
    void run(in string app_name,
             in long verbosity,
             out long retval) raises (SALOME::SALOME_Exception);
  };

  interface FSI_MILIEU: Engines::Superv_Component
  {
    void inter_run(in long NBPDTM,
                   in long NBSSIT,
                   in long ISYNCP,
                   in long NTCHR,
                   in double DTREF,
                   in double TTINIT,
                   in double EPSILO) raises (SALOME::SALOME_Exception);
  };
};

#endif
